import logging
import time
from django.urls import reverse
from oic import rndstr
from oic.oic import Client
from oic.oic.message import (
    AuthorizationResponse,
    ProviderConfigurationResponse,
    RegistrationResponse,
)
from oic.utils.authn.client import CLIENT_AUTHN_METHOD
from pretix.base.auth import BaseAuthBackend
from pretix.settings import config

logger = logging.getLogger(__name__)


class OIDCAuthBackend(BaseAuthBackend):
    def __init__(self):
        try:
            self.title = config.get(
                "oidc", "title", fallback="Login with OpenID connect"
            )

            op_info = ProviderConfigurationResponse(
                version="1.0",
                issuer=config.get("oidc", "issuer"),
                authorization_endpoint=config.get("oidc", "authorization_endpoint"),
                token_endpoint=config.get("oidc", "token_endpoint"),
                userinfo_endpoint=config.get("oidc", "userinfo_endpoint"),
                end_session_endpoint=config.get("oidc", "end_session_endpoint"),
                jwks_uri=config.get("oidc", "jwks_uri"),
            )

            client_reg = RegistrationResponse(
                client_id=config.get("oidc", "client_id"),
                client_secret=config.get("oidc", "client_secret"),
            )

            self.client = Client(client_authn_method=CLIENT_AUTHN_METHOD)
            self.client.handle_provider_config(op_info, op_info["issuer"])
            self.client.store_registration_info(client_reg)
            self.client.redirect_uris = [None]

            self.scopes = config.get("oidc", "scopes", fallback="openid").split(",")
        except KeyError:
            logger.error(
                "Please specify issuer, authorization_endpoint, token_endpoint, userinfo_endpoint, end_session_endpoint, jwks_uri, client_id and client_secret "
                "in [oidc] section in pretix.cfg"
            )

    @property
    def identifier(self):
        return "pretix_oidc"

    @property
    def verbose_name(self):
        return self.title

    def authentication_url(self, request):
        oidc_state = rndstr()
        request.session["oidc_state"] = {
            oidc_state: {
                "next": request.GET.get("next", None),
                "generated_on": int(time.time()),
            }
        }

        auth_req = self.client.construct_AuthorizationRequest(
            request_args={
                "client_id": self.client.client_id,
                "response_type": "code",
                "scope": self.scopes,
                "redirect_uri": self.redirect_uri(request),
                "state": oidc_state,
            }
        )

        return auth_req.request(self.client.authorization_endpoint)

    def redirect_uri(self, request):
        return request.build_absolute_uri(reverse("plugins:pretix_oidc:oidc_callback"))

    def get_next_url(self, request):
        return request.session.pop("oidc_next_url", None)

    def process_callback(self, request):
        auth_response = self.client.parse_response(
            AuthorizationResponse,
            info=request.META["QUERY_STRING"],
            sformat="urlencoded",
        )

        request.session["oidc_next_url"] = None
        oidc_state = request.session.pop("oidc_state", None)
        response_state = auth_response.get("state", None)

        if not oidc_state or not response_state:
            return [None, None]

        if response_state not in oidc_state:
            return [None, None]

        if oidc_state[response_state]["generated_on"] < time.time() - 5 * 60:
            return [None, None]

        request.session["oidc_next_url"] = oidc_state[response_state]["next"]

        access_token_response = self.client.do_access_token_request(
            state=auth_response["state"],
            scope=self.scopes,
            request_args={
                "code": auth_response["code"],
                "redirect_uri": self.redirect_uri(request),
            },
            authn_method="client_secret_basic",
        )

        userinfo = self.client.do_user_info_request(state=auth_response["state"])

        id_token = access_token_response["id_token"]
        user_data = {
            "uuid": userinfo[config.get("oidc", "unique_attribute", fallback="sub")],
            "email": userinfo["email"],
            "fullname": userinfo["name"],
            "auth_backend": self.identifier,
        }

        return [user_data, id_token]



# Generated by Django 3.0.10 on 2020-09-19 20:30

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("pretix_oidc", "0001_initial"),
    ]

    operations = [
        migrations.AddConstraint(
            model_name="oidcteamassignmentrule",
            constraint=models.UniqueConstraint(
                fields=("team", "attribute", "value"), name="unique_rule"
            ),
        ),
    ]



from django.dispatch import receiver
from django.urls import reverse
from django.utils.translation import gettext_lazy
from pretix.control.permissions import organizer_permission_required
from pretix.control.signals import nav_organizer


@organizer_permission_required("can_change_organizer_settings")
@receiver(nav_organizer)
def add_team_auto_assign_to_nav_pane(sender, request, **kwargs):
    """This signal is used to add the 'Team assignment rules' column to the
    navigation pane."""
    return [
        {
            "label": gettext_lazy("Team assignment rules"),
            "url": reverse(
                "plugins:pretix_oidc:team_assignment_rules",
                kwargs={"organizer": request.organizer.slug},
            ),
            "active": (
                request.resolver_match.url_name.startswith("team_assignment_rules")
            ),
            "icon": "group",
        },
    ]



# Generated by Django 3.0.8 on 2020-07-22 22:09

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("pretixbase", "0162_remove_seat_name"),
    ]

    operations = [
        migrations.CreateModel(
            name="OIDCTeamAssignmentRule",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False
                    ),
                ),
                ("attribute", models.CharField(max_length=255)),
                ("value", models.CharField(max_length=255)),
                (
                    "team",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="pretixbase.Team",
                    ),
                ),
            ],
            options={
                "verbose_name": "Team assignment rule",
            },
        ),
    ]



from dictlib import dig_get
from django.contrib import messages
from django.core.exceptions import ObjectDoesNotExist
from django.shortcuts import redirect
from django.urls import reverse
from django.utils.translation import gettext as _
from django.views.generic import TemplateView
from django.views.generic.edit import CreateView
from pretix.base.models import Team, User
from pretix.base.models.auth import EmailAddressTakenError
from pretix.control.permissions import OrganizerPermissionRequiredMixin
from pretix.control.views.auth import process_login
from pretix.helpers.compat import CompatDeleteView
from pretix.settings import config

from .auth import OIDCAuthBackend  # NOQA
from .forms import OIDCAssignmentRuleForm
from .models import OIDCTeamAssignmentRule


def oidc_callback(request):
    auth_backend = OIDCAuthBackend()
    user_data, id_token = auth_backend.process_callback(request)

    if user_data is None:
        return redirect("auth.login")

    try:
        user = User.objects.get_or_create_for_backend(
            auth_backend.identifier,
            user_data["uuid"],
            user_data["email"],
            set_always={"fullname": user_data["fullname"]},
            set_on_creation={},
        )
    except EmailAddressTakenError:
        messages.error(
            request,
            _(
                "We cannot create your user account as a user account in this system "
                "already exists with the same email address."
            ),
        )
        return redirect(reverse("control:auth.login"))
    else:
        _add_user_to_teams(user, id_token)
        _add_user_to_staff(user, id_token)
        return process_login(request, user, False)


def _add_user_to_teams(user, id_token):
    rules = OIDCTeamAssignmentRule.objects.all()

    for rule in rules:
        values = _get_attr(id_token, rule.attribute)

        if rule.value in values:
            try:
                rule.team.members.add(user)
            except ObjectDoesNotExist:
                pass
        else:
            rule.team.members.remove(user)


def _add_user_to_staff(user, id_token):
    staff_scope = config.get("oidc", "staff_scope")
    staff_values = [v.strip() for v in config.get("oidc", "staff_value").split(",")]
    if staff_scope is not None and staff_values is not None:
        values = _get_attr(id_token, staff_scope)

        user.is_staff = len(set(values) & set(staff_values)) > 0
        user.save()


def _get_attr(id_token, attr_name):
    values = dig_get(id_token, attr_name, [])
    if type(values) is not list:
        values = [values]
    return values


# These views have been adapted from pretix-cas plugin (https://github.com/DataManagementLab/pretix-cas)
class AssignmentRulesList(TemplateView, OrganizerPermissionRequiredMixin):
    template_name = "pretix_oidc/oidc_assignment_rules.html"
    permission = "can_change_organizer_settings"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        organizer = self.request.organizer
        context["teams"] = Team.objects.filter(organizer=organizer)
        context["assignmentRules"] = OIDCTeamAssignmentRule.objects.filter(
            team__organizer=organizer
        )
        return context


class AssignmentRuleEditMixin(OrganizerPermissionRequiredMixin):
    model = OIDCTeamAssignmentRule
    permission = "can_change_organizer_settings"

    def get_success_url(self):
        return reverse(
            "plugins:pretix_oidc:team_assignment_rules",
            kwargs={"organizer": self.request.organizer.slug},
        )


class AssignmentRuleUpdateMixin(AssignmentRuleEditMixin):
    fields = ["team", "attribute"]
    template_name = "pretix_oidc/oidc_assignment_rule_edit.html"

    def get_form(self, form_class=None):
        return OIDCAssignmentRuleForm(
            organizer=self.request.organizer, **self.get_form_kwargs()
        )

    def form_valid(self, form):
        super().form_valid(form)
        messages.success(self.request, _("The new assignment rule has been created."))
        return redirect(self.get_success_url())

    def form_invalid(self, form):
        messages.error(self.request, _("The assignment rule could not be created."))
        return super().form_invalid(form)


class AssignmentRuleCreate(AssignmentRuleUpdateMixin, CreateView):
    pass


class AssignmentRuleDelete(AssignmentRuleEditMixin, CompatDeleteView):
    template_name = "pretix_oidc/oidc_assignment_rule_delete.html"



from django.forms import ModelForm

from .models import OIDCTeamAssignmentRule


class OIDCAssignmentRuleForm(ModelForm):
    def __init__(self, organizer, *args, **kwargs):
        super(OIDCAssignmentRuleForm, self).__init__(*args, **kwargs)
        self.fields["team"].queryset = self.fields["team"].queryset.filter(
            organizer=organizer
        )

    class Meta:
        model = OIDCTeamAssignmentRule
        fields = ["team", "attribute", "value"]



from django.urls import path

from . import views

urlpatterns = [
    path("oidc/callback/", views.oidc_callback, name="oidc_callback"),
    path(
        "control/organizer/<str:organizer>/teams/assignment_rules/",
        views.AssignmentRulesList.as_view(),
        name="team_assignment_rules",
    ),
    path(
        "control/organizer/<str:organizer>/teams/assignment_rules/add",
        views.AssignmentRuleCreate.as_view(),
        name="team_assignment_rules.add",
    ),
    path(
        "control/organizer/<str:organizer>/teams/assignment_rules/<int:pk>/delete",
        views.AssignmentRuleDelete.as_view(),
        name="team_assignment_rules.delete",
    ),
]



from django.db import models
from django.utils.translation import gettext_lazy
from pretix.base.models import Team


class OIDCTeamAssignmentRule(models.Model):
    team = models.ForeignKey(Team, on_delete=models.CASCADE)
    attribute = models.CharField(
        max_length=255, verbose_name=gettext_lazy("OIDC attribute")
    )
    value = models.CharField(
        max_length=255, verbose_name=gettext_lazy("Attribute value")
    )

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=["team", "attribute", "value"], name="unique_rule"
            )
        ]
        verbose_name = gettext_lazy("Team assignment rule")
